"""Messenge processors."""
import os
from abc import ABC, abstractmethod
from email.message import EmailMessage
from typing import Any, Dict, Iterable, Set, Tuple

from oscrypto.asymmetric import Certificate, PrivateKey, load_certificate, load_private_key
from smail.sign import sign_message
from sqlalchemy import or_
from sqlalchemy.orm import class_mapper
from sqlalchemy.sql import func

from .backends.letter import LetterMessage
from .backends.sms import SmsMessage
from .database import atomic
from .exceptions import MessageCanceled
from .models import BlockedAddress, BlockedEmail, BlockedPhone
from .utils import normalize_email_address


class Processor(ABC):
    """Abstract base class for processors."""

    @abstractmethod
    def process_message(self, message: Any) -> Any:
        """Process and return the message.

        Raises:
            ValueError: Message can't be created.
            RuntimeError: Transient error in message creation.
            MessageCanceled: The message is canceled.
        """


class SmimeEmailProcessor(Processor):
    """Processor which signs email messages using S/MIME signatures."""

    def __init__(self, smime_certificates: Iterable[Tuple[os.PathLike, os.PathLike]]):
        # Mapping of email to loaded key and certificate.
        self.signers = self._parse_certificates(smime_certificates)

    def _parse_certificates(self, certificates: Iterable[Tuple[os.PathLike, os.PathLike]],
                            ) -> Dict[str, Tuple[PrivateKey, Certificate]]:
        """Parse keys and certificates and return mapping of normalized email address to the key-certificate pair."""
        result = {}  # type: Dict[str, Tuple[PrivateKey, Certificate]]
        for keyfile, certfile in certificates:
            # Force Path to str, since oscrypto doesn't handle Path yet.
            key = load_private_key(str(keyfile))
            cert = load_certificate(str(certfile))
            # According to RFC 8550 https://tools.ietf.org/html/rfc8550#section-3,
            # only a subjectAltName should be used as source for email addresses.
            if cert.asn1.subject_alt_name_value is None:
                continue
            for name in cert.asn1.subject_alt_name_value.native:
                normalized = normalize_email_address(name)
                if normalized is not None:
                    result[normalized] = (key, cert)
        return result

    def process_message(self, message: EmailMessage) -> EmailMessage:
        """Sign email message if possible."""
        signed_email = normalize_email_address(message['From'])
        if signed_email:
            signers = self.signers.get(signed_email)
            if signers is not None:
                message = sign_message(message, signers[0], signers[1].asn1)
        return message


class VcardEmailProcessor(Processor):
    """Processor which appends a vCard to the email messages."""

    def __init__(self, data: str):
        self.data = data

    def process_message(self, message: EmailMessage) -> EmailMessage:
        """Append vCard to the email message."""
        # Someday we may use 'text/vcard' mime type, but currently it changes behaviour, at least in Thunderbird.
        message.add_attachment(self.data, subtype='x-vcard')
        return message


class BlockedEmailProcessor(Processor):
    """Processor prevent emails from being sent to blocked email addresses.

    Removes blocked email addresses from Destination Address Fields (To, Cc, Bcc).
    If no addresses remain, the message is canceled.
    """

    address_headers = ('To', 'Cc', 'Bcc')

    def process_message(self, message: EmailMessage) -> EmailMessage:
        """Check recipients and disable blocked ones.

        Raises:
            MessageCanceled: All recipients are blocked.
        """
        # Collect all recipient addresses.
        addresses: Set[str] = set()
        for header in self.address_headers:
            if message[header]:
                addresses.update(a.addr_spec for a in message[header].addresses)

        # Find blocked recipients
        query_filter = or_(BlockedEmail.email.ilike(a) for a in addresses)
        with atomic() as session:
            blocked = tuple(b.email.lower() for b in session.query(BlockedEmail).filter(query_filter))

        # Filter out blocked recipients.
        for header in self.address_headers:
            if message[header]:
                filtered = tuple(a for a in message[header].addresses if a.addr_spec.lower() not in blocked)
                if not filtered:
                    del message[header]
                else:
                    message.replace_header(header, filtered)

        # Check if any recipients remained.
        if all(message[h] is None for h in self.address_headers):
            raise MessageCanceled("All recipients blocked.")

        return message


class BlockedSmsProcessor(Processor):
    """Processor prevent SMSes from being sent to blocked phone numbers.

    If phone is blocked, the message is canceled.
    """

    def process_message(self, message: SmsMessage) -> SmsMessage:
        """Check recipients and cancel send if phone is blocked.

        Raises:
            MessageCanceled: Phone number is blocked.
        """
        with atomic() as session:
            if session.query(BlockedPhone).filter_by(phone=message.recipient).first():
                raise MessageCanceled("Phone number blocked.")
        return message


class BlockedLetterProcessor(Processor):
    """Processor that prevents letters from being sent to blocked addresses.

    If address is blocked, the message is canceled.
    """

    attributes = tuple(k for k in class_mapper(BlockedAddress).columns.keys() if k != 'id')

    def process_message(self, message: LetterMessage) -> LetterMessage:
        """Check recipient and cancel send if address is blocked.

        Raises:
            MessageCanceled: Address is blocked.
        """
        with atomic() as session:
            query = session.query(BlockedAddress)
            for attr in self.attributes:
                value = getattr(message.recipient, attr)
                if value is not None:
                    value = value.lower()
                query = query.filter(func.lower(getattr(BlockedAddress, attr)) == value)
            if query.first():
                raise MessageCanceled("Address blocked.")
        return message
