from email.message import EmailMessage
from functools import partial
from typing import Callable, Dict, List, Sequence, Union, cast
from unittest import TestCase, skipIf
from unittest.mock import ANY, Mock, call, patch, sentinel
from uuid import UUID

from click.testing import CliRunner, Result
from filed import FileClient, Storage
from testfixtures import Comparison, LogCapture

from messenger.backends.base import BaseBackend
from messenger.backends.email import BaseEmailBackend, MemoryBackend
from messenger.backends.letter import LetterMessage
from messenger.backends.sms import SmsMessage
from messenger.commands.sender import SenderManager, _send, main
from messenger.constants import MessageStatus, TransportMethod
from messenger.database import get_session_cls
from messenger.models import Email, Letter, Sms
from messenger.settings import SETTINGS

from .utils import StaticFilter, make_email, make_letter, make_sms, override_settings

RECIPIENT = 'kryten@example.com'
SUBJECT_TEMPLATE = 'subject.txt'
BODY_TEMPLATE = 'body.txt'
SUBJECT_UUID = UUID(int=255)
BODY_UUID = UUID(int=1024)


class SenderManagerTest(TestCase):
    def test_init_empty(self):
        with override_settings(senders={}):
            manager = SenderManager()

        self.assertEqual(manager.senders, {})

    def test_init_email(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        self.assertEqual(manager.senders, {TransportMethod.EMAIL: {'memory': Comparison(MemoryBackend)}})
        sender = cast(BaseEmailBackend, manager.senders[TransportMethod.EMAIL]['memory'])
        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)
        self.assertIsNone(sender.storage)

    def test_init_email_storage(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        with override_settings(senders={'memory': sender_settings}, fileman_netloc='file.example.org'):
            manager = SenderManager()

        self.assertEqual(manager.senders, {TransportMethod.EMAIL: {'memory': Comparison(MemoryBackend)}})
        sender = cast(BaseEmailBackend, manager.senders[TransportMethod.EMAIL]['memory'])
        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)
        self.assertEqual(cast(Storage, sender.storage).client.netloc, 'file.example.org')

    def test_init_no_backend(self):
        with override_settings(senders={'memory': {}}):
            with self.assertRaisesRegex(ValueError, "'memory' has no backend defined."):
                SenderManager()

    def test_init_error(self):
        with override_settings(senders={'memory': {'backend': 'invalid'}}):
            with self.assertRaisesRegex(ValueError, "Error in loading sender memory:"):
                SenderManager()

    def test_init_filters(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url},
                           'filters': [{'class': 'messenger.tests.utils.StaticFilter'}]}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        self.assertEqual(manager.senders, {TransportMethod.EMAIL: {'memory': Comparison(MemoryBackend)}})
        sender = cast(BaseEmailBackend, manager.senders[TransportMethod.EMAIL]['memory'])
        self.assertEqual(len(sender.filters), 1)
        self.assertIsInstance(sender.filters[0], StaticFilter)

    def test_init_filters_args(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url},
                           'filters': [{'class': 'messenger.tests.test_commands_sender.StaticFilter',
                                        'result': sentinel.result}]}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        self.assertEqual(manager.senders, {TransportMethod.EMAIL: {'memory': Comparison(MemoryBackend)}})
        sender = cast(BaseEmailBackend, manager.senders[TransportMethod.EMAIL]['memory'])
        self.assertEqual(len(sender.filters), 1)
        self.assertIsInstance(sender.filters[0], StaticFilter)
        self.assertEqual(cast(StaticFilter, sender.filters[0]).result, sentinel.result)

    def test_init_filter_no_class(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url},
                           'filters': [{}]}
        with override_settings(senders={'memory': sender_settings}):
            with self.assertRaisesRegex(ValueError, "Filter has no class defined."):
                SenderManager()

    def test_init_filter_error(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url},
                           'filters': [{'class': 'invalid'}]}
        with override_settings(senders={'memory': sender_settings}):
            with self.assertRaisesRegex(ValueError, "Error in loading filter:"):
                SenderManager()

    def test_get(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        sender = cast(BaseEmailBackend, manager.get(TransportMethod.EMAIL, 'memory'))

        self.assertIsInstance(sender, MemoryBackend)
        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)
        self.assertIsNone(sender.storage)

    def test_get_sender_default(self):
        # Test first sender of the type is returned as default.
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        sender_settings_2 = {'backend': 'messenger.backends.email.MemoryBackend',
                             'options': {'secretary_url': sentinel.secretary_url_2}}
        with override_settings(senders={'memory': sender_settings, 'other': sender_settings_2}):
            manager = SenderManager()

        sender = cast(BaseEmailBackend, manager.get(TransportMethod.EMAIL))

        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)

    def test_get_unknown(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        with self.assertRaisesRegex(KeyError, "'unknown' sender not found."):
            manager.get(TransportMethod.EMAIL, 'unknown')

    def test_get_unknown_default(self):
        with override_settings(senders={}):
            manager = SenderManager()

        with self.assertRaisesRegex(KeyError, "No senders found for emails."):
            manager.get(TransportMethod.EMAIL)

    def test_getitem(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        with override_settings(senders={'memory': sender_settings}):
            manager = SenderManager()

        senders = manager[TransportMethod.EMAIL]

        self.assertEqual(len(senders), 1)
        sender = cast(BaseEmailBackend, tuple(senders)[0])
        self.assertIsInstance(sender, MemoryBackend)
        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)
        self.assertIsNone(sender.storage)

    def test_getitem_multiple(self):
        sender_settings = {'backend': 'messenger.backends.email.MemoryBackend',
                           'options': {'secretary_url': sentinel.secretary_url}}
        sender_settings_2 = {'backend': 'messenger.backends.email.MemoryBackend',
                             'options': {'secretary_url': sentinel.secretary_url_2}}
        with override_settings(senders={'memory': sender_settings, 'other': sender_settings_2}):
            manager = SenderManager()

        senders = manager[TransportMethod.EMAIL]

        self.assertEqual(len(senders), 2)
        sender = cast(BaseEmailBackend, tuple(senders)[0])
        self.assertIsInstance(sender, MemoryBackend)
        self.assertEqual(sender.name, 'memory')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url)
        sender = cast(BaseEmailBackend, tuple(senders)[1])
        self.assertIsInstance(sender, MemoryBackend)
        self.assertEqual(sender.name, 'other')
        self.assertEqual(sender.secretary.server_url, sentinel.secretary_url_2)

    def test_getitem_empty(self):
        with override_settings(senders={}):
            manager = SenderManager()

        self.assertEqual(tuple(manager[TransportMethod.EMAIL]), ())


class TestBackend(BaseBackend):
    """Simple backend for testing."""

    method = TransportMethod.EMAIL

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mock = Mock(name='backend')

    def get_message(self, model):
        return sentinel.message

    def open(self):
        self.mock.open()

    def _send(self, message):
        self.mock(message)


@override_settings(db_connection=SETTINGS['test_db_connection'])
class SendTest(TestCase):
    def setUp(self):
        self.session = get_session_cls()()
        self.log_handler = LogCapture('messenger', propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()
        self.session.close()

    def test_send_email(self):
        uuid = UUID(int=42)
        backend = TestBackend()
        make_email(self.session, uuid=uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   body_template=BODY_TEMPLATE)

        self.assertTrue(_send(TransportMethod.EMAIL, (Email.uuid == uuid), [backend]))

        # Check email was updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.SENT)
        self.assertEqual(email.attempts, 1)
        # Check backend was called
        self.assertEqual(backend.mock.mock_calls, [call.open(), call(sentinel.message)])

    def test_send_sms(self):
        uuid = UUID(int=42)
        backend = TestBackend()
        make_sms(self.session, uuid=uuid, recipient='+12345', body_template=BODY_TEMPLATE)

        self.assertTrue(_send(TransportMethod.SMS, (Sms.uuid == uuid), [backend]))

        # Check SMS was updated
        sms = self.session.query(Sms).one()
        self.assertEqual(sms.status, MessageStatus.SENT)
        self.assertEqual(sms.attempts, 1)
        # Check backend was called
        self.assertEqual(backend.mock.mock_calls, [call.open(), call(sentinel.message)])

    def test_send_letter(self):
        uuid = UUID(int=42)
        backend = TestBackend()
        make_letter(self.session, uuid=uuid, file=UUID(int=1024))

        self.assertTrue(_send(TransportMethod.LETTER, (Letter.uuid == uuid), [backend]))

        # Check letter was updated
        letter = self.session.query(Letter).one()
        self.assertEqual(letter.status, MessageStatus.SENT)
        self.assertEqual(letter.attempts, 1)
        # Check backend was called
        self.assertEqual(backend.mock.mock_calls, [call.open(), call(sentinel.message)])

    def test_failed(self):
        uuid = UUID(int=42)
        backend = TestBackend()
        backend.mock.side_effect = RuntimeError('Gazpacho!')
        make_email(self.session, uuid=uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   body_template=BODY_TEMPLATE)

        self.assertTrue(_send(TransportMethod.EMAIL, (Email.uuid == uuid), [backend]))

        # Check email was updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.PENDING)
        self.assertEqual(email.attempts, 1)
        # Check backend was called
        self.assertEqual(backend.mock.mock_calls, [call.open(), call(sentinel.message)])

    def test_skipped(self):
        # Test message skipped by filters.
        uuid = UUID(int=42)
        filter = Mock(spec=['filter'])
        filter.filter.return_value = False
        backend = TestBackend(filters=[filter])
        make_email(self.session, uuid=uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   body_template=BODY_TEMPLATE)

        self.assertFalse(_send(TransportMethod.EMAIL, (Email.uuid == uuid), [backend]))

        # Check email was not updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.PENDING)
        self.assertEqual(email.attempts, 0)
        # Check backend was not called
        self.assertEqual(backend.mock.mock_calls, [])
        # Check warning
        msg = ("Message <Email 00000000-0000-0000-0000-00000000002a to ['kryten@example.com']> is skipped, because "
               "it didn't match any filters.")
        self.log_handler.check(('messenger.commands.sender', 'WARNING', msg))

    def test_not_found(self):
        # Test message not found.
        backend = TestBackend()

        self.assertFalse(_send(TransportMethod.EMAIL, (Email.uuid == Email.uuid), [backend]))

        # Check backend was not called
        self.assertEqual(backend.mock.mock_calls, [])

    @skipIf(SETTINGS['test_db_connection'].startswith('sqlite'), 'Does not work on SQLite database.')
    def test_locked(self):
        # Test message is locked.
        uuid = UUID(int=42)
        backend = TestBackend()
        make_email(self.session, uuid=uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   body_template=BODY_TEMPLATE)

        # Lock email
        lock_session = get_session_cls()()
        lock_session.query(Email).filter_by(uuid=uuid).with_for_update().all()

        self.assertFalse(_send(TransportMethod.EMAIL, (Email.uuid == uuid), [backend]))

        # Release the lock, so the data can be checked.
        lock_session.rollback()

        # Check email was not updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.PENDING)
        self.assertEqual(email.attempts, 0)
        # Check backend was not called
        self.assertEqual(backend.mock.mock_calls, [])


@override_settings(db_connection=SETTINGS['test_db_connection'])
class MainTest(TestCase):
    senders = {'email': {'backend': 'messenger.backends.email.MemoryBackend',
                         'options': {'secretary_url': 'http://secretary.example/api/'}},
               'sms': {'backend': 'messenger.backends.sms.MemorySmsBackend',
                       'options': {'secretary_url': 'http://secretary.example/api/'}},
               'letter': {'backend': 'messenger.backends.letter.MemoryLetterBackend'}}

    def setUp(self):
        patcher = patch('messenger.backends.base.SecretaryClient', autospec=True)
        self.addCleanup(patcher.stop)
        client_mock = patcher.start()
        self.secretary_mock = client_mock.return_value

        self.session = get_session_cls()()
        self.runner = CliRunner(mix_stderr=False)
        MemoryBackend.outbox.clear()

        self.file_client = Mock(FileClient)
        self.storage = Storage(self.file_client)
        storage_patcher = patch('messenger.commands.sender._get_storage', autospec=True, return_value=self.storage)
        self.addCleanup(storage_patcher.stop)
        storage_patcher.start()

        self.log_handler = LogCapture('messenger', propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()
        self.session.close()

    def test_main_no_emails(self):
        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [])

    def test_main_storage(self):
        # Test run with storage set up.
        with override_settings(senders=self.senders, fileman_netloc='fileman.example'):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [])

    def test_main_sent(self):
        make_email(self.session, recipient=[RECIPIENT], status=MessageStatus.SENT, subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [])

    def _test_main_email(self, args: Sequence[str]):
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        self.runner.invoke(main, args)

        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage), Comparison(EmailMessage)])
        self.assertEqual(MemoryBackend.outbox[0]['To'], RECIPIENT)

        # Check emails were updated
        emails = self.session.query(Email)
        self.assertEqual([e.status for e in emails], [MessageStatus.SENT, MessageStatus.SENT])

    def test_main_email(self):
        with override_settings(senders=self.senders):
            self._test_main_email([])

    def test_main_sms(self):
        self.secretary_mock.render.return_value = ''
        recipient = '+123456'
        make_sms(self.session, recipient=recipient, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_sms(self.session, recipient=recipient, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [Comparison(SmsMessage), Comparison(SmsMessage)])
        self.assertEqual(MemoryBackend.outbox[0].recipient, recipient)

        # Check SMS were updated
        sms = self.session.query(Sms)
        self.assertEqual([m.status for m in sms], [MessageStatus.SENT, MessageStatus.SENT])

    def test_main_letter(self):
        file_uuid = UUID(int=1024)
        make_letter(self.session, file=file_uuid)
        make_letter(self.session, file=file_uuid)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [Comparison(LetterMessage), Comparison(LetterMessage)])
        self.assertEqual(MemoryBackend.outbox[0].file.uuid, file_uuid)

        # Check letters were updated
        letters = self.session.query(Letter)
        self.assertEqual([m.status for m in letters], [MessageStatus.SENT, MessageStatus.SENT])

    def _test_main_all(self, args: List):
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_sms(self.session, recipient='+123456', body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_letter(self.session, file=UUID(int=1024))

        with override_settings(senders=self.senders):
            self.runner.invoke(main, args)

    def test_main_all(self):
        self._test_main_all([])
        self.assertEqual(MemoryBackend.outbox,
                         [Comparison(EmailMessage), Comparison(SmsMessage), Comparison(LetterMessage)])

    def test_main_email_only(self):
        self._test_main_all(['--type', 'email'])
        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage)])

    def test_main_sms_only(self):
        self._test_main_all(['--type', 'sms'])
        self.assertEqual(MemoryBackend.outbox, [Comparison(SmsMessage)])

    def test_main_letter_only(self):
        self._test_main_all(['--type', 'letter'])
        self.assertEqual(MemoryBackend.outbox, [Comparison(LetterMessage)])

    def test_main_custom_sender(self):
        senders = {'custom': {'backend': 'messenger.backends.email.MemoryBackend',
                              'options': {'secretary_url': 'http://secretary.example/api/'}}}
        with override_settings(senders=senders):
            self._test_main_email(['--sender', 'custom', '--type', 'email'])

    def test_main_failed(self):
        # Test emails, which failed to be sent, are not processed multiple times in one run.
        self.secretary_mock.render.side_effect = RuntimeError
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        self.assertEqual(MemoryBackend.outbox, [])

        # Check emails were updated
        emails = self.session.query(Email)
        self.assertEqual([e.status for e in emails], [MessageStatus.PENDING, MessageStatus.PENDING])
        self.assertEqual([e.attempts for e in emails], [1, 1])

    @skipIf(SETTINGS['test_db_connection'].startswith('sqlite'), 'Does not work on SQLite database.')
    def test_main_lock(self):
        # Test locked emails are skipped.
        self.secretary_mock.render.return_value = ''
        uuid = UUID(int=1)
        locked_uuid = UUID(int=42)
        make_email(self.session, uuid=uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        make_email(self.session, uuid=locked_uuid, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        # Lock email
        lock_session = get_session_cls()()
        lock_session.query(Email).filter_by(uuid=locked_uuid).with_for_update().all()

        with override_settings(senders=self.senders):
            self.runner.invoke(main, [])

        # Release the lock, so the data can be checked.
        lock_session.rollback()

        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage)])

        # Check emails were updated
        emails = self.session.query(Email).order_by(Email.uuid).filter(Email.uuid.in_([uuid, locked_uuid]))
        self.assertEqual([e.status for e in emails], [MessageStatus.SENT, MessageStatus.PENDING])
        self.assertEqual([e.attempts for e in emails], [1, 0])

    def test_main_retry_failed(self):
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], status=MessageStatus.FAILED, subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, ['--retry-failed'])

        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage)])
        self.assertEqual(MemoryBackend.outbox[0]['To'], RECIPIENT)

        # Check email was updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.SENT)

    def test_main_retry_error(self):
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], status=MessageStatus.ERROR, subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=self.senders):
            self.runner.invoke(main, ['--retry-errors'])

        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage)])
        self.assertEqual(MemoryBackend.outbox[0]['To'], RECIPIENT)

        # Check email was updated
        email = self.session.query(Email).one()
        self.assertEqual(email.status, MessageStatus.SENT)

    def test_main_loop(self):
        # Test the loop mode.
        # Hack the time.sleep to trigger custom events.
        self.secretary_mock.render.return_value = ''
        email_maker = partial(
            make_email, self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
            subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)
        side_effects = [email_maker, KeyboardInterrupt()]  # type: List[Union[Callable, BaseException]]

        def call_effects(*args):
            effect = side_effects.pop(0)
            if isinstance(effect, BaseException):
                raise effect
            return effect()

        with patch('messenger.commands.sender.sleep') as sleep_mock:
            sleep_mock.side_effect = call_effects

            make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                       subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

            with override_settings(senders=self.senders):
                self.runner.invoke(main, ['--loop'])

        self.assertEqual(MemoryBackend.outbox, [Comparison(EmailMessage), Comparison(EmailMessage)])
        self.assertEqual(MemoryBackend.outbox[0]['To'], RECIPIENT)

        # Check emails were updated
        emails = self.session.query(Email)
        self.assertEqual([e.status for e in emails], [MessageStatus.SENT, MessageStatus.SENT])
        # Check sleep was called
        self.assertEqual(sleep_mock.mock_calls, [call(300), call(300)])

    def test_main_loop_delay(self):
        # Test the loop delay option.
        # Hack the time.sleep to trigger custom events.
        with patch('messenger.commands.sender.sleep') as sleep_mock:
            sleep_mock.side_effect = [None, KeyboardInterrupt]
            with override_settings(senders=self.senders):
                self.runner.invoke(main, ['--loop', '--loop-delay', '1024'])

        # Check sleep was called
        self.assertEqual(sleep_mock.mock_calls, [call(1024), call(1024)])

    def test_main_connections(self):
        # Test the connection opening and closing.
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        sender_mock = Mock(BaseBackend)
        with patch('messenger.commands.sender._get_senders') as get_senders_mock:
            get_senders_mock.return_value = [sender_mock]

            self.runner.invoke(main, ['--type', 'email'])

        # Check calls to senders.
        self.assertEqual(sender_mock.mock_calls, [call.filter(ANY), call.open(), call.send(ANY), call.close()])

    def test_main_connections_loop(self):
        # Test the connection opening and closing in loop mode.
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        sender_mock = Mock(BaseBackend)
        with patch('messenger.commands.sender._get_senders') as get_senders_mock:
            # Use the patch to return sender twice and interrupt the process on third run.
            get_senders_mock.side_effect = [[sender_mock], [sender_mock], KeyboardInterrupt]

            # Set delay to 0 to skip sleep between cycles.
            self.runner.invoke(main, ['--type', 'email', '--loop', '--loop-delay', '0'])

        # Check calls to senders.
        calls = [call.filter(ANY), call.open(), call.send(ANY), call.close()] * 2
        self.assertEqual(sender_mock.mock_calls, calls)

    def _test_output(self, verbosity: int, senders: Dict = senders) -> Result:
        self.secretary_mock.render.return_value = ''
        make_email(self.session, recipient=[RECIPIENT], subject_template=SUBJECT_TEMPLATE,
                   subject_template_uuid=SUBJECT_UUID, body_template=BODY_TEMPLATE, body_template_uuid=BODY_UUID)

        with override_settings(senders=senders):
            result = self.runner.invoke(main, ['--verbosity', str(verbosity), '--type', 'email'])

        return result

    def test_silent(self):
        result = self._test_output(0)
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.stderr, '')

    def test_info(self):
        result = self._test_output(1)
        self.assertEqual(result.stdout, 'Emails: 1 sent, 0 skipped.\n')
        self.assertEqual(result.stderr, '')

    def test_detail(self):
        result = self._test_output(2)
        self.assertEqual(result.stdout, '.\nEmails: 1 sent, 0 skipped.\n')
        self.assertEqual(result.stderr, '')

    def _test_skipped_output(self, verbosity: int) -> Result:
        senders = {'smtp': {'backend': 'messenger.backends.email.MemoryBackend',
                            'filters': [{'class': 'messenger.tests.utils.StaticFilter', 'result': False}],
                            'options': {'secretary_url': 'http://secretary.example/api/'}}}
        return self._test_output(verbosity, senders)

    def test_silent_skipped(self):
        result = self._test_skipped_output(0)
        self.assertEqual(result.stdout, '')
        self.assertEqual(result.stderr, '')

    def test_info_skipped(self):
        result = self._test_skipped_output(1)
        self.assertEqual(result.stdout, 'Emails: 0 sent, 1 skipped.\n')
        self.assertEqual(result.stderr, '')

    def test_detail_skipped(self):
        result = self._test_skipped_output(2)
        self.assertEqual(result.stdout, 's\nEmails: 0 sent, 1 skipped.\n')
        self.assertEqual(result.stderr, '')

    def test_main_no_senders(self):
        with override_settings(senders={}):
            result = self.runner.invoke(main)

        self.assertNotEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, '')
        self.assertIn("No senders found for emails.", result.stderr)

    def test_main_settings_error(self):
        with override_settings(senders={'memory': {}}):
            result = self.runner.invoke(main)

        self.assertNotEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, '')
        self.assertIn("'memory' has no backend defined.", result.stderr)

    def test_main_unknown_sender(self):
        with override_settings(senders={}):
            result = self.runner.invoke(main, ['--sender', 'unknown', '--type', 'email'])

        self.assertNotEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, '')
        self.assertIn("No senders found for emails.", result.stderr)

    def test_sender_all_types(self):
        with override_settings(senders={}):
            result = self.runner.invoke(main, ['--sender', 'unknown'])

        self.assertNotEqual(result.exit_code, 0)
        self.assertEqual(result.stdout, '')
        self.assertIn("Sender can only be used with single type.", result.stderr)
