from datetime import datetime
from email.message import EmailMessage, MIMEPart
from typing import Any, Dict, Type
from unittest import TestCase
from unittest.mock import call, patch, sentinel
from uuid import UUID

from freezegun import freeze_time
from imap_tools import AND, MailMessage
from pytz import UTC

from messenger.constants import ArchiveMessageStatus, MessageStatus
from messenger.database import get_session_cls
from messenger.delivery.email import MIMETYPE_DSN, MIMETYPE_MESSAGE, MIMETYPE_REPORT, REPORT_TYPE_DSN, ImapDelivery
from messenger.models import ArchiveEmail, Email
from messenger.settings import SETTINGS

from .utils import make_archive_email, make_email, override_settings

try:
    from imap_tools import MailMessageFlags
except ImportError:  # pragma: no cover
    # Fallback for imap_tools < 0.18
    from imap_tools import MessageFlags as MailMessageFlags


@freeze_time('2020-02-01 14:00:00')
@override_settings(db_connection=SETTINGS['test_db_connection'])
class ImapDeliveryTest(TestCase):
    recipient = 'kryten@example.org'
    subject_template = 'subject.txt'
    body_template = 'body.txt'
    message_id = '<original@example.org>'

    reply_id = '<reply@example.org>'
    reply_subject = 'Quagaars!'
    reply_date = datetime(2001, 2, 3, 4, tzinfo=UTC)

    def setUp(self):
        self.session = get_session_cls()()

    def tearDown(self):
        self.session.close()

    def get_message(self, headers: Dict[str, Any] = None) -> EmailMessage:
        message = EmailMessage()
        message['Subject'] = self.reply_subject
        message['Message-ID'] = self.reply_id
        for key, value in (headers or {}).items():
            message[key] = value
        return message

    def get_delivery_status(self, message_dsn_fields: Dict[str, Any] = None,
                            recipient_dsn_fields: Dict[str, Any] = None) -> MIMEPart:
        delivery_status = MIMEPart()
        delivery_status['Content-Type'] = MIMETYPE_DSN
        message_dsn = MIMEPart()
        for key, value in (message_dsn_fields or {}).items():
            message_dsn[key] = value
        delivery_status.attach(message_dsn)
        recipient_dsn = MIMEPart()
        for key, value in (recipient_dsn_fields or {}).items():
            recipient_dsn[key] = value
        delivery_status.attach(recipient_dsn)
        return delivery_status

    def get_returned_message(self) -> MIMEPart:
        returned_message = MIMEPart()
        returned_message['Content-Type'] = MIMETYPE_MESSAGE
        returned_message['Message-ID'] = self.message_id
        original_message = MIMEPart()
        returned_message.attach(original_message)
        return returned_message

    def get_dsn_message(self, headers: Dict[str, Any] = None, delivery_status: MIMEPart = None,
                        returned_message: MIMEPart = None) -> EmailMessage:
        headers = headers or {}
        headers['Content-Type'] = '{}; report-type={}'.format(MIMETYPE_REPORT, REPORT_TYPE_DSN)
        message = self.get_message(headers)
        if delivery_status:
            message.attach(delivery_status)
        if returned_message:
            message.attach(returned_message)
        return message

    def test_get_mailbox_default(self):
        # Test mailbox created with default arguments.
        delivery = ImapDelivery()

        with patch('ssl.create_default_context', autospec=True, return_value=sentinel.context):
            with patch('messenger.delivery.email.MailBox', autospec=True, return_value=sentinel.mailbox) as imap_mock:
                self.assertEqual(delivery.get_mailbox(), sentinel.mailbox)

        self.assertEqual(imap_mock.mock_calls, [call(host='', port=None, ssl_context=sentinel.context)])

    def test_get_mailbox_insecure(self):
        # Test mailbox created with default arguments.
        delivery = ImapDelivery(ssl=False)

        with patch('messenger.delivery.email.MailBoxUnencrypted') as imap_mock:
            imap_mock.return_value = sentinel.mailbox

            self.assertEqual(delivery.get_mailbox(), sentinel.mailbox)

        self.assertEqual(imap_mock.mock_calls, [call(host='', port=None)])

    def test_get_mailbox_custom(self):
        # Test mailbox created with custom arguments.
        delivery = ImapDelivery(host=sentinel.host, port=sentinel.port, ssl=False)

        with patch('messenger.delivery.email.MailBoxUnencrypted', return_value=sentinel.mailbox) as imap_mock:
            self.assertEqual(delivery.get_mailbox(), sentinel.mailbox)

        self.assertEqual(imap_mock.mock_calls, [call(host=sentinel.host, port=sentinel.port)])

    def test_get_mailbox_ssl(self):
        data = (
            # certfile, keyfile, ssl_calls
            (None, None, [call()]),
            (sentinel.certfile, None, [call(), call().load_cert_chain(sentinel.certfile, None)]),
            (sentinel.certfile, sentinel.keyfile,
             [call(), call().load_cert_chain(sentinel.certfile, sentinel.keyfile)]),
        )
        for certfile, keyfile, ssl_calls in data:
            with self.subTest(certfile=certfile, keyfile=keyfile):
                delivery = ImapDelivery(host=sentinel.host, port=sentinel.port, keyfile=keyfile, certfile=certfile)

                with patch('ssl.create_default_context', autospec=True) as ssl_mock:
                    with patch('messenger.delivery.email.MailBox', autospec=True,
                               return_value=sentinel.mailbox) as imap_mock:
                        self.assertEqual(delivery.get_mailbox(), sentinel.mailbox)

                self.assertEqual(imap_mock.mock_calls,
                                 [call(host=sentinel.host, port=sentinel.port, ssl_context=ssl_mock.return_value)])
                self.assertEqual(ssl_mock.mock_calls, ssl_calls)

    def test_get_mailbox_tls(self):
        data = (
            # certfile, keyfile, ssl_calls
            (None, None, [call()]),
            (sentinel.certfile, None, [call(), call().load_cert_chain(sentinel.certfile, None)]),
            (sentinel.certfile, sentinel.keyfile,
             [call(), call().load_cert_chain(sentinel.certfile, sentinel.keyfile)]),
        )
        for certfile, keyfile, ssl_calls in data:
            with self.subTest(certfile=certfile, keyfile=keyfile):
                delivery = ImapDelivery(ssl=False, tls=True, certfile=certfile, keyfile=keyfile)

                with patch('ssl.create_default_context', autospec=True) as ssl_mock:
                    # Skip autospec=True because it doesn't match box property.
                    with patch('messenger.delivery.email.MailBoxUnencrypted') as imap_mock:
                        mailbox = delivery.get_mailbox()

                self.assertEqual(imap_mock.mock_calls,
                                 [call(host='', port=None), call().box.starttls(ssl_mock.return_value)])
                self.assertEqual(ssl_mock.mock_calls, ssl_calls)
                # Check calls first, then mailbox. To avoid __eq__ call on mailbox.
                self.assertEqual(mailbox, imap_mock.return_value)

    def test_get_mailbox_login(self):
        # Test login. Ensure tls is started before login.
        delivery = ImapDelivery(ssl=False, tls=True, username=sentinel.username, password=sentinel.password)

        with patch('ssl.create_default_context', autospec=True, return_value=sentinel.context):
            # Skip autospec=True because it doesn't match box property.
            with patch('messenger.delivery.email.MailBoxUnencrypted') as imap_mock:
                delivery.get_mailbox()

        calls = [call(host='', port=None),
                 call().box.starttls(sentinel.context),
                 call().login(username=sentinel.username, password=sentinel.password)]
        self.assertEqual(imap_mock.mock_calls, calls)

    def test_is_dsn(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message()))

        self.assertFalse(delivery.is_dsn(message))

    def test_is_dsn_report(self):
        # Test message is a report, but not a delivery status
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message({'Content-Type': MIMETYPE_REPORT})))

        self.assertFalse(delivery.is_dsn(message))

    def test_is_dsn_delivery(self):
        # Test message is a report, but not a delivery status
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_dsn_message()))

        self.assertTrue(delivery.is_dsn(message))

    def test_get_status(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message()))

        self.assertEqual(delivery.get_status(message), MessageStatus.DELIVERED)

    def test_get_status_report(self):
        # Test message is a report, but not a delivery status
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message({'Content-Type': MIMETYPE_REPORT})))

        self.assertIsNone(delivery.get_status(message))

    def test_get_status_delivery(self):
        # Test message is a delivery status, but does not contain `Action` field
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_dsn_message()))

        self.assertIsNone(delivery.get_status(message))

    def test_get_status_delivery_failed(self):
        # Test message is a delivery status with `Action` failed
        delivery = ImapDelivery()
        delivery_status = self.get_delivery_status(recipient_dsn_fields={'Action': 'failed'})
        message = MailMessage.from_bytes(bytes(self.get_dsn_message(delivery_status=delivery_status)))

        self.assertEqual(delivery.get_status(message), MessageStatus.UNDELIVERED)

    def test_get_status_delivery_delivered(self):
        # Test message is a delivery status with `Action` delivered
        delivery = ImapDelivery()
        delivery_status = self.get_delivery_status(recipient_dsn_fields={'Action': 'delivered'})
        message = MailMessage.from_bytes(bytes(self.get_dsn_message(delivery_status=delivery_status)))

        self.assertEqual(delivery.get_status(message), MessageStatus.DELIVERED)

    def test_get_status_delivery_other(self):
        # Test message is a delivery status with `Action` failed
        delivery = ImapDelivery()
        delivery_status = self.get_delivery_status(recipient_dsn_fields={'Action': 'delayed'})
        message = MailMessage.from_bytes(bytes(self.get_dsn_message(delivery_status=delivery_status)))

        self.assertIsNone(delivery.get_status(message))

    def test_get_message_id_none(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message()))

        self.assertIsNone(delivery.get_message_id(message))

    def test_get_message_id_reply(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message({'In-Reply-To': self.message_id})))

        self.assertEqual(delivery.get_message_id(message), self.message_id)

    def test_get_message_id_references(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message({'References': self.message_id})))

        self.assertEqual(delivery.get_message_id(message), self.message_id)

    def test_get_message_id_original_message(self):
        # Test Message-ID extracted from original message in DSN.
        delivery = ImapDelivery()
        returned_message = self.get_returned_message()
        message = MailMessage.from_bytes(bytes(self.get_dsn_message(returned_message=returned_message)))

        self.assertEqual(delivery.get_message_id(message), self.message_id)

    def test_get_dsn_fields_empty(self):
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_dsn_message()))

        self.assertEqual(delivery.get_dsn_fields(message), {})

    def test_get_dsn_fields_all(self):
        delivery = ImapDelivery()
        message_dsn_fields = {'Reporting-MTA': 'dns; mail.example.org'}
        recipient_dsn_fields = {'Action': 'failed', 'Status': '5.1.1', 'Remote-MTA': 'dns; mx1.example.org',
                                'Diagnostic-Code': 'smtp; 550 5.1.1 Sorry, no mailbox here by that name.',
                                'Final-Recipient': 'rfc822; {}'.format(self.recipient)}
        delivery_status = self.get_delivery_status(message_dsn_fields=message_dsn_fields,
                                                   recipient_dsn_fields=recipient_dsn_fields)
        message = MailMessage.from_bytes(bytes(self.get_dsn_message(delivery_status=delivery_status)))

        self.assertEqual(delivery.get_dsn_fields(message), dict(**message_dsn_fields, **recipient_dsn_fields))

    def test_handle_message_unknown(self):
        # Test message with no identifiers does nothing.
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(self.get_message()))

        delivery.handle_message(message)

        self.assertEqual(self.session.query(Email).count(), 0)

    def test_handle_message_unspecific(self):
        # Test message, which doesn't provide information about delivery.
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        delivery = ImapDelivery()
        message = self.get_message({'In-Reply-To': self.message_id, 'Content-Type': MIMETYPE_REPORT})
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(Email).one()
        # Check email wasn't updated
        self.assertEqual(email.status, MessageStatus.SENT)
        self.assertIsNone(email.update_datetime)
        self.assertIsNone(email.delivery_datetime)
        self.assertIsNone(email.delivery_info)

    def test_handle_message_processed(self):
        # Test email is not updated if in other status
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.UNDELIVERED,
                   message_id=self.message_id, subject_template=self.subject_template, body_template=self.body_template)
        delivery = ImapDelivery()
        message = self.get_message({'In-Reply-To': self.message_id})
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(Email).one()
        # Check email wasn't updated
        self.assertEqual(email.status, MessageStatus.UNDELIVERED)
        self.assertIsNone(email.update_datetime)
        self.assertIsNone(email.delivery_datetime)
        self.assertIsNone(email.delivery_info)

    def test_handle_message_processed_archive(self):
        # Test archive email is not updated if in other status
        make_archive_email(
            self.session, id=42, create_datetime=datetime(1999, 8, 7, 6, tzinfo=UTC), recipient=[self.recipient],
            status=ArchiveMessageStatus.UNDELIVERED, message_id=self.message_id,
            subject_template=self.subject_template, subject_template_uuid=UUID(int=255),
            body_template=self.body_template, body_template_uuid=UUID(int=256))
        delivery = ImapDelivery()
        message = self.get_message({'In-Reply-To': self.message_id})
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(ArchiveEmail).one()
        # Check email wasn't updated
        self.assertEqual(email.status, MessageStatus.UNDELIVERED)
        self.assertIsNone(email.update_datetime)
        self.assertIsNone(email.delivery_datetime)
        self.assertIsNone(email.delivery_info)

    def test_handle_message_headless(self):
        # Message with no useful headers.
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        delivery = ImapDelivery()
        message = EmailMessage()
        message['In-Reply-To'] = self.message_id
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(Email).one()
        # Check email was updated
        self.assertEqual(email.status, MessageStatus.DELIVERED)
        self.assertEqual(email.update_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_info, {})

    def _test_handle_message_delivered(self, message: EmailMessage, email_cls: Type,
                                       delivery_datetime=datetime(2020, 2, 1, 14, tzinfo=UTC)):
        # If the message references known Message-ID mark as delivered.
        delivery = ImapDelivery()
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(email_cls).one()
        # Check email was updated
        self.assertEqual(email.status, MessageStatus.DELIVERED)
        self.assertEqual(email.update_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_datetime, delivery_datetime)
        self.assertEqual(email.delivery_info, {'Message-ID': self.reply_id, 'Subject': self.reply_subject})

    def test_handle_message_delivered(self):
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        message = self.get_message({'In-Reply-To': self.message_id})
        self._test_handle_message_delivered(message, Email)

    def test_handle_message_delivered_archive(self):
        make_archive_email(
            self.session, id=42, create_datetime=datetime(1999, 8, 7, 6, tzinfo=UTC), recipient=[self.recipient],
            message_id=self.message_id, subject_template=self.subject_template, subject_template_uuid=UUID(int=255),
            body_template=self.body_template, body_template_uuid=UUID(int=256))
        message = self.get_message({'In-Reply-To': self.message_id})
        self._test_handle_message_delivered(message, ArchiveEmail)

    def test_handle_message_delivered_date(self):
        # Test Date header is used as a delivery time
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        message = self.get_message({'In-Reply-To': self.message_id, 'Date': self.reply_date})
        self._test_handle_message_delivered(message, Email, delivery_datetime=self.reply_date)

    def _test_handle_message_undelivered(
            self, email_cls: Type, headers: Dict[str, Any] = None, returned_message: MIMEPart = None,
            delivery_datetime=datetime(2020, 2, 1, 14, tzinfo=UTC)):
        # If the message is undelivered delivery status.
        message_dsn_fields = {'Reporting-MTA': 'dns; mail.example.org'}
        recipient_dsn_fields = {'Action': 'failed', 'Status': '5.1.1'}
        delivery = ImapDelivery()
        delivery_status = self.get_delivery_status(message_dsn_fields=message_dsn_fields,
                                                   recipient_dsn_fields=recipient_dsn_fields)
        message = self.get_dsn_message(headers, delivery_status=delivery_status,
                                       returned_message=returned_message)
        message = MailMessage.from_bytes(bytes(message))

        delivery.handle_message(message)

        email = self.session.query(email_cls).one()
        # Check email was updated
        self.assertEqual(email.status, MessageStatus.UNDELIVERED)
        self.assertEqual(email.update_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_datetime, delivery_datetime)
        delivery_info = dict({'Message-ID': self.reply_id, 'Subject': self.reply_subject},
                             **message_dsn_fields, **recipient_dsn_fields)
        self.assertEqual(email.delivery_info, delivery_info)

    def test_handle_message_undelivered(self):
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        self._test_handle_message_undelivered(Email, {'In-Reply-To': self.message_id})

    def test_handle_message_undelivered_archive(self):
        make_archive_email(
            self.session, id=42, create_datetime=datetime(1999, 8, 7, 6, tzinfo=UTC), recipient=[self.recipient],
            message_id=self.message_id, subject_template=self.subject_template, subject_template_uuid=UUID(int=255),
            body_template=self.body_template, body_template_uuid=UUID(int=256))
        self._test_handle_message_undelivered(ArchiveEmail, {'In-Reply-To': self.message_id})

    def test_handle_message_undelivered_date(self):
        # Test Date header is used as a delivery time
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        self._test_handle_message_undelivered(Email, {'In-Reply-To': self.message_id, 'Date': self.reply_date},
                                              delivery_datetime=self.reply_date)

    def test_check_delivery_empty(self):
        delivery = ImapDelivery(ssl=False)

        with patch('messenger.delivery.email.MailBoxUnencrypted', autospec=True) as imap_mock:
            imap_mock.return_value.__enter__.return_value.fetch.return_value = []

            self.assertEqual(delivery.check_delivery(), 0)

        calls = [call(host='', port=None),
                 call().__enter__(),
                 call().__enter__().fetch(AND(seen=False, deleted=False), mark_seen=False, limit=None),
                 call().__enter__().expunge(),
                 call().__exit__(None, None, None)]
        self.assertEqual(imap_mock.mock_calls, calls)

    def test_check_delivery(self):
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        delivery = ImapDelivery(ssl=False)
        message = self.get_message({'In-Reply-To': self.message_id})
        message = MailMessage([(b'UID 42', bytes(message))])

        with patch('messenger.delivery.email.MailBoxUnencrypted', autospec=True) as imap_mock:
            imap_mock.return_value.__enter__.return_value.fetch.return_value = [message]

            self.assertEqual(delivery.check_delivery(), 1)

        email = self.session.query(Email).one()
        # Check email was updated
        self.assertEqual(email.status, MessageStatus.DELIVERED)
        self.assertEqual(email.update_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_info, {'Message-ID': self.reply_id, 'Subject': self.reply_subject})

        calls = [call(host='', port=None),
                 call().__enter__(),
                 call().__enter__().fetch(AND(seen=False, deleted=False), mark_seen=False, limit=None),
                 call().__enter__().flag('42', [MailMessageFlags.SEEN], True),
                 call().__enter__().expunge(),
                 call().__exit__(None, None, None)]
        self.assertEqual(imap_mock.mock_calls, calls)

    def test_check_delivery_undelivered(self):
        make_email(self.session, recipient=[self.recipient], status=MessageStatus.SENT, message_id=self.message_id,
                   subject_template=self.subject_template, body_template=self.body_template)
        delivery = ImapDelivery(ssl=False)
        delivery_status = self.get_delivery_status(recipient_dsn_fields={'Action': 'failed'})
        message = self.get_dsn_message({'In-Reply-To': self.message_id}, delivery_status=delivery_status)
        message = MailMessage([(b'UID 42', bytes(message))])

        with patch('messenger.delivery.email.MailBoxUnencrypted', autospec=True) as imap_mock:
            imap_mock.return_value.__enter__.return_value.fetch.return_value = [message]

            self.assertEqual(delivery.check_delivery(), 1)

        email = self.session.query(Email).one()
        # Check email was updated
        self.assertEqual(email.status, MessageStatus.UNDELIVERED)
        self.assertEqual(email.update_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_datetime, datetime(2020, 2, 1, 14, tzinfo=UTC))
        self.assertEqual(email.delivery_info,
                         {'Message-ID': self.reply_id, 'Subject': self.reply_subject, 'Action': 'failed'})

        calls = [call(host='', port=None),
                 call().__enter__(),
                 call().__enter__().fetch(AND(seen=False, deleted=False), mark_seen=False, limit=None),
                 call().__enter__().flag('42', [MailMessageFlags.SEEN, MailMessageFlags.DELETED], True),
                 call().__enter__().expunge(),
                 call().__exit__(None, None, None)]
        self.assertEqual(imap_mock.mock_calls, calls)

    def test_check_delivery_limit(self):
        delivery = ImapDelivery(ssl=False)

        with patch('messenger.delivery.email.MailBoxUnencrypted', autospec=True) as imap_mock:
            imap_mock.return_value.__enter__.return_value.fetch.return_value = []

            self.assertEqual(delivery.check_delivery(limit=sentinel.limit), 0)

        calls = [call(host='', port=None),
                 call().__enter__(),
                 call().__enter__().fetch(AND(seen=False, deleted=False), mark_seen=False, limit=sentinel.limit),
                 call().__enter__().expunge(),
                 call().__exit__(None, None, None)]
        self.assertEqual(imap_mock.mock_calls, calls)
