import os
from email.message import EmailMessage
from email.mime.base import MIMEBase
from pathlib import Path
from typing import Tuple, cast
from unittest import TestCase
from unittest.mock import sentinel
from uuid import UUID

from oscrypto.asymmetric import Certificate, PrivateKey
from testfixtures import Comparison

from messenger.backends.letter import LetterMessage
from messenger.backends.sms import SmsMessage
from messenger.database import get_session_cls
from messenger.exceptions import MessageCanceled
from messenger.models import Address, BlockedAddress, BlockedEmail, BlockedPhone
from messenger.processors import (BlockedEmailProcessor, BlockedLetterProcessor, BlockedSmsProcessor,
                                  SmimeEmailProcessor, VcardEmailProcessor)
from messenger.settings import SETTINGS

from .utils import override_settings

CERTIFICATE_PATH = Path(os.path.dirname(__file__)) / 'certificates'


class VcardEmailProcessorTest(TestCase):
    def test_process_message(self):
        message = EmailMessage()
        vcard = 'VCARD DATA\n'
        processor = VcardEmailProcessor(vcard)

        result = processor.process_message(message)

        attachments = cast(Tuple[EmailMessage], tuple(result.iter_attachments()))
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0].get_content_type(), 'text/x-vcard')
        self.assertEqual(attachments[0].get_content(), vcard)


class SmimeEmailProcessorTest(TestCase):
    sender = 'holly@example.org'
    recipient = 'kryten@example.org'
    subject = 'Secret message'
    content = 'Gazpacho!'

    def test_parse_certificates_no_certs(self):
        # Test _parse_certificates with no certificates defined.
        processor = SmimeEmailProcessor([])
        self.assertEqual(processor.signers, {})

    def test_get_signers_no_mails(self):
        # Test _parse_certificates with certificate with no emails
        keyfile = CERTIFICATE_PATH / 'test_ca.key'
        certfile = CERTIFICATE_PATH / 'test_ca.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        self.assertEqual(processor.signers, {})

    def test_parse_certificates_invalid_email(self):
        # Test _parse_certificates with certificates with invalid email.
        keyfile = CERTIFICATE_PATH / 'holly.key'
        certfile = CERTIFICATE_PATH / 'holly-invalid.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        self.assertEqual(processor.signers, {})

    def test_parse_certificates(self):
        # Test _parse_certificates with certificate with valid email.
        keyfile = CERTIFICATE_PATH / 'holly.key'
        certfile = CERTIFICATE_PATH / 'holly.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        self.assertEqual(processor.signers, {'holly@example.org': (Comparison(PrivateKey), Comparison(Certificate))})

    def test_parse_certificates_bare(self):
        # Test _parse_certificates with certificate with valid email.
        keyfile = CERTIFICATE_PATH / 'holly.key'
        certfile = CERTIFICATE_PATH / 'holly-expanded.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        self.assertEqual(processor.signers, {'holly@example.org': (Comparison(PrivateKey), Comparison(Certificate))})

    def _test_process_message_signed(self, sender: str):
        keyfile = CERTIFICATE_PATH / 'holly.key'
        certfile = CERTIFICATE_PATH / 'holly.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        message = EmailMessage()
        message['From'] = sender
        message['To'] = self.recipient
        message['Subject'] = self.subject
        message.set_content(self.content)

        signed_message = processor.process_message(message)

        self.assertEqual(signed_message.get_content_type(), 'multipart/signed')
        self.assertEqual(signed_message['From'], sender)
        self.assertEqual(signed_message['To'], self.recipient)
        self.assertEqual(signed_message['Subject'], self.subject)
        body, signature = cast(Tuple[EmailMessage, MIMEBase], signed_message.get_payload())
        self.assertEqual(body.get_content().strip(), self.content)
        self.assertEqual(signature.get_content_type(), 'application/pkcs7-signature')

    def test_get_message_signed(self):
        self._test_process_message_signed('holly@example.org')

    def test_get_message_signed_bare(self):
        # Test message is signed with matching email address - bare email.
        self._test_process_message_signed('Holly <holly@example.org>')

    def test_get_message_not_signed(self):
        # Test message is not signed, if sender doesn't match certificate.
        keyfile = CERTIFICATE_PATH / 'holly.key'
        certfile = CERTIFICATE_PATH / 'holly.crt'
        processor = SmimeEmailProcessor([(keyfile, certfile)])
        sender = 'queeg-500@example.org'
        message = EmailMessage()
        message['From'] = sender
        message['To'] = self.recipient
        message['Subject'] = self.subject
        message.set_content(self.content)

        unsigned_message = processor.process_message(message)

        self.assertEqual(unsigned_message.get_content_type(), 'text/plain')
        self.assertEqual(unsigned_message['From'], sender)
        self.assertEqual(unsigned_message['To'], self.recipient)
        self.assertEqual(unsigned_message['Subject'], self.subject)
        self.assertEqual(unsigned_message.get_content().strip(), self.content)


@override_settings(db_connection=SETTINGS['test_db_connection'])
class BlockedEmailProcessorTest(TestCase):
    def setUp(self):
        self.session = get_session_cls()()
        self.processor = BlockedEmailProcessor()

    def tearDown(self):
        self.session.close()

    def get_message(self, recipient: str):
        message = EmailMessage()
        message['To'] = recipient
        return message

    def block_email(self, email: str):
        self.session.add(BlockedEmail(email=email))
        self.session.commit()

    def test_no_blocked(self):
        recipient = 'kryten@example.org'
        message = self.get_message(recipient)

        result = self.processor.process_message(message)

        self.assertEqual(result['To'], recipient)

    def test_blocked(self):
        recipient = 'kryten@example.org, rimmer@example.org'
        self.block_email('rimmer@example.org')
        message = self.get_message(recipient)

        result = self.processor.process_message(message)

        self.assertEqual(result['To'], 'kryten@example.org')

    def _test_canceled(self, recipient: str, blocked: str):
        self.block_email(blocked)
        message = self.get_message(recipient)

        with self.assertRaises(MessageCanceled):
            self.processor.process_message(message)

    def test_canceled(self):
        self._test_canceled('kryten@example.org', 'kryten@example.org')

    def test_canceled_case_insensitive(self):
        self._test_canceled('KRYTEN@example.org', 'kryten@EXAMPLE.org')

    def test_canceled_name(self):
        self._test_canceled('Kryten <kryten@example.org>', 'kryten@example.org')

    def test_canceled_unicode(self):
        self._test_canceled('Kryten 🤖 <kryten-🤖@example.org>', 'kryten-🤖@example.org')

    def _test_blocked_copy_headers(self, header: str):
        recipient = 'kryten@example.org'
        blocked = 'rimmer@example.org'
        self.block_email(blocked)
        message = self.get_message(recipient)
        message[header] = blocked

        result = self.processor.process_message(message)

        self.assertEqual(result['To'], 'kryten@example.org')
        self.assertIsNone(result[header])

    def test_blocked_cc(self):
        self._test_blocked_copy_headers('Cc')

    def test_blocked_bcc(self):
        self._test_blocked_copy_headers('Bcc')


@override_settings(db_connection=SETTINGS['test_db_connection'])
class BlockedSmsProcessorTest(TestCase):
    def setUp(self):
        self.session = get_session_cls()()
        self.processor = BlockedSmsProcessor()

    def tearDown(self):
        self.session.close()

    def test_no_blocked(self):
        recipient = '+1234'
        message = SmsMessage(UUID(int=42), recipient, '')

        result = self.processor.process_message(message)

        self.assertEqual(result.recipient, recipient)

    def test_canceled(self):
        recipient = '+1234'
        self.session.add(BlockedPhone(phone=recipient))
        self.session.commit()
        message = SmsMessage(UUID(int=42), recipient, '')

        with self.assertRaises(MessageCanceled):
            self.processor.process_message(message)


@override_settings(db_connection=SETTINGS['test_db_connection'])
class BlockedLetterProcessorTest(TestCase):
    def setUp(self):
        self.session = get_session_cls()()
        self.processor = BlockedLetterProcessor()

    def tearDown(self):
        self.session.close()

    def test_no_blocked(self):
        recipient = Address(name='Arnold Rimmer', organization='Jupiter Mining Company', street='Deck 16',
                            city='Red Dwarf', state='Deep Space', postal_code='JMC', country='Jupiter')
        message = LetterMessage(sentinel.uuid, recipient, sentinel.file)

        result = self.processor.process_message(message)

        self.assertEqual(result.recipient, recipient)

    def _test_not_blocked(self, recipient: Address, blocked: Address) -> None:
        # Test when message is not blocked.
        blocked_address = BlockedAddress(
            name=blocked.name, organization=blocked.organization, street=blocked.street, city=blocked.city,
            state=blocked.state, postal_code=blocked.postal_code, country=blocked.country)
        self.session.add(blocked_address)
        self.session.commit()
        message = LetterMessage(sentinel.uuid, recipient, sentinel.file)

        result = self.processor.process_message(message)

        self.assertEqual(result.recipient, recipient)

    def test_partial_block_not_blocked(self):
        # Test partial address is different from full address - blocked is partial.
        recipient = Address(name='Arnold Rimmer', organization='Jupiter Mining Company', street='Deck 16',
                            city='Red Dwarf', state='Deep Space', postal_code='JMC', country='Jupiter')
        blocked = Address(name='Arnold Rimmer', city='Red Dwarf')
        self._test_not_blocked(recipient, blocked)

    def test_partial_recipient_not_blocked(self):
        # Test partial address is different from full address - recipient is partial.
        recipient = Address(name='Arnold Rimmer', city='Red Dwarf')
        blocked = Address(name='Arnold Rimmer', organization='Jupiter Mining Company', street='Deck 16',
                          city='Red Dwarf', state='Deep Space', postal_code='JMC', country='Jupiter')
        self._test_not_blocked(recipient, blocked)

    def test_canceled(self):
        recipient = Address(name='Arnold Rimmer', organization='Jupiter Mining Company', street='Deck 16',
                            city='Red Dwarf', state='Deep Space', postal_code='JMC', country='Jupiter')
        blocked_address = BlockedAddress(
            name=recipient.name, organization=recipient.organization, street=recipient.street, city=recipient.city,
            state=recipient.state, postal_code=recipient.postal_code, country=recipient.country)
        self.session.add(blocked_address)
        self.session.commit()
        message = LetterMessage(sentinel.uuid, recipient, sentinel.file)

        with self.assertRaises(MessageCanceled):
            self.processor.process_message(message)

    def test_canceled_case_insensitive(self):
        recipient = Address(name='Arnold Rimmer', organization='Jupiter Mining Company', street='Deck 16',
                            city='Red Dwarf', state='Deep Space', postal_code='JMC', country='Jupiter')
        blocked = Address(name='arnold rimmer', organization='jupiter mining company', street='deck 16',
                          city='red dwarf', state='deep space', postal_code='jmc', country='jupiter')
        blocked_address = BlockedAddress(
            name=blocked.name, organization=blocked.organization, street=blocked.street, city=blocked.city,
            state=blocked.state, postal_code=blocked.postal_code, country=blocked.country)
        self.session.add(blocked_address)
        self.session.commit()
        message = LetterMessage(sentinel.uuid, recipient, sentinel.file)

        with self.assertRaises(MessageCanceled):
            self.processor.process_message(message)
