from unittest import TestCase
from unittest.mock import call, patch, sentinel

from messenger.settings import ENVIRON_NAME
from messenger.utils import (create_ssl_context, import_path, load_instance, normalize_email_address,
                             normalize_phone_number)


class NormalizeEmailAddress(TestCase):
    def test_valid(self):
        data = (
            ('holly@example.org', 'holly@example.org'),
            ('HOLLY@example.org', 'holly@example.org'),
            ('HOLLY@EXAMPLE.ORG', 'holly@example.org'),
            ('Holly <holly@example.org>', 'holly@example.org'),
            ('Holly <HOLLY@EXAMPLE.ORG>', 'holly@example.org'),
        )
        for raw, normalized in data:
            with self.subTest(raw=raw):
                self.assertEqual(normalize_email_address(raw), normalized)

    def test_invalid(self):
        data = (
            '',
            '<>',
        )
        for raw in data:
            with self.subTest(raw=raw):
                self.assertIsNone(normalize_email_address(raw))


class NormalizePhoneNumber(TestCase):
    def test_valid(self):
        data = (
            ('+1234', '+1234'),  # E164 format
            (' +1234 ', '+1234'),
            ('+1.234', '+1234'),  # E164 format used in FRED
        )
        for raw, normalized in data:
            with self.subTest(raw=raw):
                self.assertEqual(normalize_phone_number(raw), normalized)

    def test_invalid(self):
        data = (
            '',
            'invalid',
            '1234',  # Just a number, not a phone number.
        )
        for raw in data:
            with self.subTest(raw=raw):
                self.assertIsNone(normalize_phone_number(raw))


class CreateSslContextTest(TestCase):
    def test_context(self):
        data = (
            # certfile, keyfile, ssl_calls
            (None, None, [call()]),
            (sentinel.certfile, None, [call(), call().load_cert_chain(sentinel.certfile, None)]),
            (sentinel.certfile, sentinel.keyfile,
             [call(), call().load_cert_chain(sentinel.certfile, sentinel.keyfile)]),
        )
        for certfile, keyfile, ssl_calls in data:
            with self.subTest(certfile=certfile, keyfile=keyfile):
                with patch('ssl.create_default_context', autospec=True) as ssl_mock:
                    context = create_ssl_context(certfile, keyfile)

                # Check calls first, then the result. To avoid __eq__ call on mocked context.
                self.assertEqual(ssl_mock.mock_calls, ssl_calls)
                self.assertEqual(context, ssl_mock.return_value)


class ImportPathTest(TestCase):
    def test_class(self):
        cls = import_path('messenger.tests.test_utils.ImportPathTest')
        self.assertEqual(cls, ImportPathTest)

    def test_attribute(self):
        attribute = import_path('messenger.settings.ENVIRON_NAME')
        self.assertEqual(attribute, ENVIRON_NAME)

    def test_invalid(self):
        data = (
            # path, error_message
            ('', 'is not a valid dotted path'),
            ('.', 'Empty module name'),
            ('.invalid', 'Empty module name'),
            ('invalid', 'is not a valid dotted path'),
            ('invalid.', 'No module named'),
            ('invalid.path', 'No module named'),
            ('invalid..path', 'No module named'),
            ('messenger.tests.invalid.TestCase', 'No module named'),
            ('messenger.tests.test_utils.', 'does not have a class or attribute'),
            ('messenger.tests.test_utils.InvalidTestCase', 'does not have a class or attribute'),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    import_path(path)


class TestClass:
    """Test class for load_instance."""
    def __init__(self, *, keyword=None, another=None):
        self.keyword = keyword
        self.another = another


class LoadInstanceTest(TestCase):
    def test_class_only(self):
        obj = load_instance('messenger.tests.test_utils.TestClass')

        self.assertIsInstance(obj, TestClass)
        self.assertIsNone(obj.keyword)
        self.assertIsNone(obj.another)

    def test_class_kwargs(self):
        obj = load_instance('messenger.tests.test_utils.TestClass', keyword=sentinel.value, another=sentinel.another)

        self.assertIsInstance(obj, TestClass)
        self.assertEqual(obj.keyword, sentinel.value)
        self.assertEqual(obj.another, sentinel.another)

    def test_invalid_path(self):
        data = (
            # path, error_message
            ('', 'is not a valid dotted path'),
            ('messenger.tests.test_utils.InvalidTestCase', 'does not have a class or attribute'),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    load_instance(path)

    def test_invalid_kwargs(self):
        data = (
                {'invalid': sentinel.invalid},
                {'keyword': sentinel.keyword, 'another': sentinel.another, 'invalid': sentinel.invalid},
        )
        for kwargs in data:
            with self.subTest(kwargs=kwargs):
                with self.assertRaises(TypeError):
                    load_instance('messenger.tests.test_utils.TestClass', **kwargs)
